import json
from llamalogs.helpers import ms_time

class Stat:
    def __init__(self):
        self.component = ''
        self.name = ''
        self.value = 0
        self.type = ''
        self.account = ''
        self.graph = ''
        self.timestamp = ms_time()
        self.count = 0

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, 
            sort_keys=True, indent=4)

    def toAPIFormat(self):
        return self.__dict__


from llamalogs.helpers import ms_time

class Log:
    def __init__(self):
        self.sender = ''
        self.receiver = ''
        self.timestamp = ms_time()
        self.message = ''
        self.initialMessage = True
        self.account = ''
        self.graph = ''
        self.isError = False
        self.elapsed = 0


import threading 
import traceback
import schedule

from llamalogs.logAggregator import LogAggregator
from llamalogs.log import Log
from llamalogs.returnLog import ReturnLog
from llamalogs.stat import Stat
from llamalogs.llamaProxy import LlamaProxy

class LlamaLogs:
    globalAccountKey = ''
    globalGraphName = ''
    isDisabled = False
    commThread = None

    @staticmethod
    def init(options = {}):
        try:
            if ("accountKey" in options):
                LlamaLogs.globalAccountKey = str(options["accountKey"])
            if ("graphName" in options):
                LlamaLogs.globalGraphName = str(options["graphName"])
            if ("isDevEnv" in options):
                LlamaProxy.isDevEnv = options["isDevEnv"] or False
            if ("disabled" in options):
                LlamaLogs.isDisabled = options["disabled"] or False

            if LlamaLogs.commThread is None and LlamaLogs.isDisabled is False:
                LlamaLogs.commThread = threading.Thread(target=LogAggregator.start_timer)
                LlamaLogs.commThread.daemon = True
                LlamaLogs.commThread.start()
        except:
            print("`LlamaLogs Error: init function")
    
    @staticmethod
    def stop():
        try:
            if LlamaLogs.commThread is not None:
                schedule.clear()
                LlamaLogs.commThread = None
        except:
            print("`LlamaLogs Error: stop function")

    @staticmethod
    def point_stat(options = {}):
        try:
            if LlamaLogs.isDisabled: 
                return
            options["type"] = "point"
            LlamaLogs.processStat(options)
        except:
            print("LlamaLogs Error: point_stat function")

    @staticmethod
    def avg_stat(options = {}):
        try:
            if LlamaLogs.isDisabled: 
                return
            options["type"] = "average"
            LlamaLogs.processStat(options)
        except:
            print("LlamaLogs Error: avg_stat function")

    @staticmethod
    def max_stat(options = {}):
        try:
            if LlamaLogs.isDisabled: 
                return
            options["type"] = "max"
            LlamaLogs.processStat(options)
        except:
            print("LlamaLogs Error: max_stat function")

    @staticmethod
    def log(options = {}, returnLog = None):
        try:
            if LlamaLogs.isDisabled: 
                return
            return LlamaLogs.processLog(options, returnLog)
        except:
            print("LlamaLogs Error: log function")
            traceback.print_exc()

    @staticmethod
    def force_send():
        try:
            if LlamaLogs.isDisabled: 
                return
            LogAggregator.send_messages()
        except:
            print("LlamaLogs Error: force_send function")

    @staticmethod
    def processStat(options = {}):
        stat = Stat()
        stat.component = options["component"]
        stat.name = options["name"]
        stat.value = options["value"]
        stat.type = options["type"]

        if ("accountKey" in options):
            stat.account = options["accountKey"]
        else:
            stat.account = LlamaLogs.globalAccountKey or -1 

        if ("graphName" in options):
            stat.graph = options["graphName"]
        else:
            stat.graph = LlamaLogs.globalGraphName
        
        LogAggregator.add_stat(stat)

    @staticmethod
    def processLog(options = {}, returnLog = {}):
        if (returnLog is not None):
            options["s"] = returnLog.sender
            options["r"] = returnLog.receiver
            options["initialMessage"] = returnLog.initialMessage
            options["startTime"] = returnLog.startTime
            options["accountKey"] = returnLog.accountKey
            options["graphName"] = returnLog.graphName

        log = Log()
        if ("s" in options):
            log.sender = options["s"]
        else:
            log.sender = options["sender"]

        if ("r" in options):
            log.receiver = options["r"]
        else:
            log.receiver = options["receiver"]
        
        if ("message" in options):
            log.message = options["message"]

        if ("isError" in options):
            log.isError = (options["isError"] and True) or False

        if ("accountKey" in options):
            log.account = options["accountKey"]
        else:
            log.account = LlamaLogs.globalAccountKey or -1 

        if ("graphName" in options):
            log.graph = options["graphName"]
        else:
            log.graph = LlamaLogs.globalGraphName

        if ("initialMessage" in options):
            log.initialMessage = options["initialMessage"]

        if ("startTime" in options):
            log.elapsed = log.timestamp - options["startTime"]

        if (log.sender == "" or log.receiver == "" or log.graph == "" or log.account == ""):
            return None
        

        LogAggregator.add_log(log)
        return ReturnLog(log)


import schedule
import time
import json
import threading
import sys

from llamalogs.llamaProxy import LlamaProxy
from llamalogs.aggregateLog import AggregateLog

class LogAggregator:
	#lock for the aggregate log/stat maps
	lock = threading.Lock()
	aggregateLogs = {}
	aggregateStats = {}

	@staticmethod
	def start_timer():
		# sending messages after 5 seconds upon startup, then to a 1 min interval
		time.sleep(5)
		LogAggregator.send_messages()

		delay = 59.5
		schedule.every(delay).seconds.do(LogAggregator.send_messages)
		while True and len(schedule.jobs) > 0:
			schedule.run_pending()
			time.sleep(delay)

	@staticmethod
	def send_messages():
		if (LlamaProxy.isDevEnv):
			print("Sent Messages")
		log_list, stat_list = LogAggregator.gather_messages()
		LlamaProxy.send_messages(log_list, stat_list)

	@staticmethod
	def gather_messages():
		with LogAggregator.lock:
			currentLogs = LogAggregator.aggregateLogs
			LogAggregator.aggregateLogs = {}
			currentStats = LogAggregator.aggregateStats
			LogAggregator.aggregateStats = {}

		log_list = []
		stat_list = []
		for sender in currentLogs:
			for receiver in currentLogs[sender]:
				log_list.append(currentLogs[sender][receiver].toAPIFormat())
				
		for component in currentStats:
			for name in currentStats[component]:
				stat_list.append(currentStats[component][name].toAPIFormat())
		
		return log_list, stat_list

	@staticmethod
	def add_log(log):
		with LogAggregator.lock:
			if (log.sender not in LogAggregator.aggregateLogs):
				LogAggregator.aggregateLogs[log.sender] = {}
			if (log.receiver not in LogAggregator.aggregateLogs[log.sender]):
				LogAggregator.aggregateLogs[log.sender][log.receiver] = AggregateLog(log)

			working_ob = LogAggregator.aggregateLogs[log.sender][log.receiver]

			if (log.isError):
				working_ob.errors = working_ob.errors + 1
			if (log.elapsed):
				prev_amount = working_ob.elapsed * working_ob.elapsedCount
				working_ob.elapsed = (prev_amount + log.elapsed) / (working_ob.total + 1)
				working_ob.elapsedCount = working_ob.elapsedCount + 1
			if (log.initialMessage):
				working_ob.initialMessageCount = working_ob.initialMessageCount + 1

			working_ob.total = working_ob.total + 1
			if (working_ob.message == '' and log.isError == False):
				working_ob.message = str(log.message or '')
			if (working_ob.errorMessage == '' and log.isError == True):
				working_ob.errorMessage = str(log.message or '')
		if (LlamaProxy.isDevEnv):
			print("adding log")
			print(LogAggregator.aggregateLogs)

	@staticmethod
	def add_stat(stat):
		with LogAggregator.lock:
			if (stat.type == "point"):
				if (stat.component not in LogAggregator.aggregateStats):
					LogAggregator.aggregateStats[stat.component] = {}
				LogAggregator.aggregateStats[stat.component][stat.name] = stat    

			if (stat.type == "average"):
				LogAggregator.add_stat_avg(stat)
			if (stat.type == "max"):
				LogAggregator.add_stat_max(stat)

	@staticmethod
	def add_stat_avg(stat):
		if (stat.component not in LogAggregator.aggregateStats):
			LogAggregator.aggregateStats[stat.component] = {}
		if (stat.name not in LogAggregator.aggregateStats[stat.component]):
			LogAggregator.aggregateStats[stat.component][stat.name] = stat
			stat.count = 0

		existing = LogAggregator.aggregateStats[stat.component][stat.name]
		existing.value = existing.value + stat.value
		existing.count = existing.count + 1
	
	@staticmethod
	def add_stat_max(stat):
		if (stat.component not in LogAggregator.aggregateStats):
			LogAggregator.aggregateStats[stat.component] = {}
		if (stat.name not in LogAggregator.aggregateStats[stat.component]):
			LogAggregator.aggregateStats[stat.component][stat.name] = stat

		existing = LogAggregator.aggregateStats[stat.component][stat.name]
		if (stat.value > existing.value):
			existing.value = stat.value



import json
from llamalogs.helpers import ms_time

class AggregateLog:
    def __init__(self, log):
        self.sender = log.sender
        self.receiver = log.receiver
        self.account = log.account
        self.message = ''
        self.errorMessage = ''
        self.initialMessageCount = 0
        self.graph = log.graph
        self.total = 0
        self.errors = 0
        self.elapsed = 0
        self.elapsedCount = 0

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, 
            sort_keys=True, indent=4)

    def toAPIFormat(self):
        api_log = {}
        api_log["sender"] = self.sender
        api_log["receiver"] = self.receiver
        api_log["count"] = self.total
        api_log["errorCount"] = self.errors
        api_log["message"] = self.message
        api_log["errorMessage"] = self.errorMessage
        api_log["clientTimestamp"] = ms_time()
        api_log["graph"] = self.graph or 'noGraph'
        api_log["account"] = self.account
        api_log["initialMessageCount"] = self.initialMessageCount
        return api_log


import requests

class LlamaProxy:
	isDevEnv = False

	@staticmethod
	def send_messages(currentLogs, currentStats):
		LlamaProxy.send_data(currentLogs, currentStats)

	@staticmethod
	def send_data(log_list, stat_list):
		url = 'https://llamalogs.com/'
		if (LlamaProxy.isDevEnv):
			url = 'http://localhost:4000/'

		if (LlamaProxy.isDevEnv):
			print("data_lists")
			print(log_list)
			print(stat_list)	

		if (len(log_list) or len(stat_list)):
			try:
				first_log = (len(log_list) and log_list[0]) or (len(stat_list) and stat_list[0])
				account_key = first_log["account"]
				# connect and receive timouts
				requests.post(url + 'api/v0/timedata', json = {"account_key": account_key, "time_logs": log_list, "time_stats": stat_list}, timeout=(2, 5))
			except:
				print('LlamaLogs Error; error sending data to llama logs server')	


class ReturnLog:
    def __init__(self, log):
        self.sender = log.receiver
        self.receiver = log.sender
        self.initialMessage = False
        self.startTime = log.timestamp
        self.accountKey = log.account
        self.graphName = log.graph


from distutils.core import setup

setup(
  name = 'llamalogs',   
  packages = ['llamalogs'],
  version = '0.1.3',
  license='MIT',    
  description = 'Client library for Llama Logs; https://llamalogs.com',
  author = 'Llama Logs',
  author_email = 'andrew@llamalogs.com',
  url = 'https://github.com/llamalogs/llamalogs-py',
  download_url = 'https://github.com/llamalogs/llamalogs-py/archive/0.1.3.tar.gz',
  keywords = ['llama', 'logs', 'metrics', 'llamalogs'],   
  install_requires=['requests', 'schedule'],
  classifiers=[
    'Development Status :: 3 - Alpha',     
    'Intended Audience :: Developers',     
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',   
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)